from typing import Any, Union, get_origin

try:
    from types import UnionType  # type: ignore[attr-defined]
except ImportError:
    UnionType = None


def is_union(tp: Any) -> bool:
    origin = get_origin(tp)
    return origin == Union or (UnionType is not None and origin == UnionType)


NoneType = type(None)


def normalize_type(tp: Any) -> Any:
    if tp == NoneType or tp is None:
        return None
    return tp



from itertools import chain
from typing import List, Optional, Sequence, TypedDict

import pydantic
from pydantic.error_wrappers import ErrorWrapper, ValidationError
from pydantic.errors import MissingError, NoneIsNotAllowedError

from vitya.errors_base import (
    ExactFieldLenError,
    IncorrectData,
    IncorrectLen,
    NeedRequiredField,
)
from vitya.payment_order.errors import DocumentNumberValidationBOEmptyNotAllowed
from vitya.payment_order.payments.checkers import CheckerError


def flatten_error_wrappers(error_list: pydantic.error_wrappers.ErrorList) -> List[ErrorWrapper]:
    if isinstance(error_list, Sequence):
        return list(chain.from_iterable(map(flatten_error_wrappers, error_list)))
    return [error_list]


class AlertKeyToFieldName(TypedDict):
    amount: str
    payer: str
    payer_bic: str
    payer_account_number: str
    receiver_bic: str
    receiver: str
    receiver_account_number: str


class AlertBody(TypedDict):
    alert: Optional[str]
    failed_field: str


class AlertGenerator:
    _key_to_ru = {
        'amount': 'Сумма',
        'payer': 'Плательщик',
        'receiver_bic': 'БИК банка получателя',
        'receiver': 'Получатель',
        'receiver_account_number': 'Номер счёта получателя',
        'payer_account_number': 'Номер счёта плательщика',
        'payer_bic': 'БИК банка плательщика',
    }

    def __init__(self, key_to_field_name: AlertKeyToFieldName):
        self._field_name_to_key = {value: key for key, value in key_to_field_name.items()}

    def _mixin_to_alert(self, exc: Exception) -> Optional[str]:
        if isinstance(exc, NeedRequiredField):
            return f'Поле «{exc.target_ru}» должно быть заполнено'
        if isinstance(exc, IncorrectLen):
            return f'Поле «{exc.target_ru}» содержит неправильное количество символов'
        if isinstance(exc, ExactFieldLenError):
            return f'Поле «{exc.target_ru}» должно содержать ровно {exc.required_len} символов'
        if isinstance(exc, IncorrectData):
            return f'Поле «{exc.target_ru}» содержит некорректные данные'
        if isinstance(exc, DocumentNumberValidationBOEmptyNotAllowed):
            return (
                'При выборе «24» в поле «101, Статус плательщика» обязательны данные'
                ' хотя бы одного из полей: «60, ИНН плательщика», «22, УИН», «108, Документ»'
            )
        return None

    def get_error_client_alerts(self, exc: Exception) -> List[AlertBody]:
        if not isinstance(exc, ValidationError):
            alert = self._mixin_to_alert(exc)
            if alert is not None:
                return [AlertBody(alert=alert, failed_field=exc.target)]  # type: ignore
            return [AlertBody(alert=None, failed_field=getattr(exc, 'target', None))]

        result = []
        for error_wrapper in flatten_error_wrappers(exc.raw_errors):
            if isinstance(error_wrapper.exc, CheckerError):
                for sub_error in error_wrapper.exc.errors:
                    alert = self._mixin_to_alert(sub_error)
                    if alert is not None:
                        result.append(AlertBody(alert=alert, failed_field=sub_error.target))  # type: ignore
            elif isinstance(error_wrapper.exc, (NoneIsNotAllowedError, MissingError)):
                if len(error_wrapper.loc_tuple()) == 1:
                    field_name = error_wrapper.loc_tuple()[0]
                    if field_name in self._field_name_to_key:
                        target_ru = self._key_to_ru[self._field_name_to_key[field_name]]
                        result.append(AlertBody(
                            alert=f'Поле «{target_ru}» должно быть заполнено',
                            failed_field=self._field_name_to_key[field_name]
                        ))
            else:
                alert = self._mixin_to_alert(error_wrapper.exc)
                if alert is not None:
                    result.append(AlertBody(alert=alert, failed_field=error_wrapper.exc.target))  # type: ignore
                else:
                    result.append(AlertBody(alert=None, failed_field=error_wrapper.exc.target))  # type: ignore
        return result



from typing import Any, Optional


class VityaDescribedError(Exception):
    target: Optional[str]
    description: str

    target_ru: Optional[str]
    description_ru: str

    def __str__(self) -> str:
        if self.target:
            return f'invalid {self.target}: {self.description}'
        return f'{self.description}'

    def __getattr__(self, item: str) -> Any:
        if item.endswith('_ru'):
            return getattr(self, item[:-3])
        raise AttributeError


class NeedRequiredField(VityaDescribedError):
    pass


class IncorrectLen(VityaDescribedError):
    pass


class ExactFieldLenError(VityaDescribedError):
    required_len: int


class IncorrectData(VityaDescribedError):
    pass



import re
from datetime import date
from decimal import Decimal, InvalidOperation
from typing import Optional, Union

from vitya.payment_order.errors import (
    AccountNumberValidationDigitsOnlyError,
    AccountNumberValidationSizeError,
    AccountNumberValidationTypeError,
    AmountNotANumber,
    AmountValidationLengthError,
    AmountValidationLessOrEqualZeroError,
    CBCValidationTypeError,
    CBCValidationValueCannotZerosOnly,
    CBCValidationValueDigitsOnlyError,
    CBCValidationValueLenError,
    CustomerValidationSizeError,
    DocumentDateValidationTypeError,
    DocumentNumberValidationTypeError,
    NumberValidationLenError,
    OperationKindValidationTypeError,
    OperationKindValidationValueError,
    PayerStatusValidationTypeError,
    PayerStatusValidationValueError,
    PayerValidationSizeError,
    PaymentOrderLenError,
    PaymentOrderValidationError,
    PurposeCodeValidationTypeError,
    PurposeValidationMaxLenError,
    PurposeValidationTypeError,
    ReasonValidationTypeError,
    ReasonValidationValueLenError,
    ReceiverAccountNumberValidationDigitsOnlyError,
    ReceiverAccountNumberValidationSizeError,
    ReceiverAccountNumberValidationTypeError,
    ReceiverValidationNameError,
    ReceiverValidationSizeError,
    TaxPeriodValidationTypeError,
    TypeOfIncomeValidationError,
    TypeOfIncomeValidationTypeError,
    UINValidationControlSumError,
    UINValidationDigitsOnlyError,
    UINValidationLenError,
    UINValidationOnlyZeroError,
    UINValidationTypeError,
)
from vitya.payment_order.payments.constants import (
    CHANGE_YEAR,
    CHARS_FOR_PURPOSE,
    PAYER_STATUSES,
    PAYER_STATUSES_AFTER_2024,
    REPLACE_CHARS_FOR_SPACE,
)


def validate_number(
    value: str,
) -> str:
    if len(value) > 6:
        raise NumberValidationLenError
    return value


def validate_amount(amount: str) -> Decimal:
    if isinstance(amount, str):
        if len(amount) > 18:
            raise AmountValidationLengthError

    try:
        value = Decimal(amount)
    except InvalidOperation:
        raise AmountNotANumber
    if value <= Decimal(0.0):
        raise AmountValidationLessOrEqualZeroError

    return value


def validate_customer(value: str) -> str:
    len_value = len(value)
    if len_value < 1 or len_value > 160:
        raise CustomerValidationSizeError
    return value


def validate_payer(value: str) -> str:
    try:
        validate_customer(value)
    except CustomerValidationSizeError as e:
        raise PayerValidationSizeError from e
    return value


def validate_receiver(value: str) -> str:
    try:
        validate_customer(value)
    except CustomerValidationSizeError as e:
        raise ReceiverValidationSizeError from e
    if bool(re.match('(.*)(4)[0-9]{19}', value)):
        raise ReceiverValidationNameError
    return value


def validate_payment_order(value: Optional[Union[int, str]]) -> int:
    if value is None or value == '':
        return 5

    if isinstance(value, str) and len(value) != 1:
        raise PaymentOrderLenError
    try:
        value_int = int(value)
    except ValueError:
        raise PaymentOrderValidationError
    if value_int not in {1, 2, 3, 4, 5}:
        raise PaymentOrderValidationError
    return value_int


def validate_account_number(value: str) -> str:
    if not isinstance(value, str):
        raise AccountNumberValidationTypeError
    if len(value) != 20:
        raise AccountNumberValidationSizeError
    if not only_digits(value):
        raise AccountNumberValidationDigitsOnlyError
    return value


def validate_receiver_account_number(value: str) -> str:
    try:
        return validate_account_number(value)
    except AccountNumberValidationTypeError:
        raise ReceiverAccountNumberValidationTypeError
    except AccountNumberValidationSizeError:
        raise ReceiverAccountNumberValidationSizeError
    except AccountNumberValidationDigitsOnlyError:
        raise ReceiverAccountNumberValidationDigitsOnlyError


def validate_operation_kind(value: str) -> str:
    if not isinstance(value, str):
        raise OperationKindValidationTypeError
    if len(value) != 2:
        raise OperationKindValidationValueError
    return value


def validate_uin_control_sum(
    value: str
) -> None:
    if not only_digits(value):
        raise UINValidationDigitsOnlyError
    if len(value) != 20 and len(value) != 25:
        return

    count = 1
    sum_ = 0
    for c in value[:-1:]:
        if count > 10:
            count = 1
        sum_ += int(c) * count
        count += 1

    if sum_ == 0:
        raise UINValidationOnlyZeroError

    mod_11 = sum_ % 11
    if mod_11 != 10:
        if mod_11 != int(value[-1]):
            raise UINValidationControlSumError
        return

    count = 3
    sum_ = 0
    for c in value[:-1:]:
        if count > 10:
            count = 1
        sum_ += int(c) * count
        count += 1
    mod_11 = sum_ % 11
    mod_11 = 0 if mod_11 == 10 else mod_11
    if mod_11 != int(value[-1]):
        raise UINValidationControlSumError


def validate_uin(value: str) -> Optional[str]:
    if not isinstance(value, str):
        raise UINValidationTypeError
    if value in {'', '0'}:
        return None
    if not (len(value) == 4 or len(value) == 20 or len(value) == 25):
        raise UINValidationLenError
    if len(value) == 4 and value == '0000':
        raise UINValidationOnlyZeroError
    validate_uin_control_sum(value)
    return value


def only_digits(value: str) -> bool:
    return re.match(r'^[0-9]+$', value) is not None


def validate_purpose_code(value: int) -> int:
    try:
        value = int(value)
    except Exception:
        raise PurposeCodeValidationTypeError
    return value


_PURPOSE_TRANS_TABLE = str.maketrans({char: ' ' for char in REPLACE_CHARS_FOR_SPACE})


def validate_purpose(value: str) -> Optional[str]:
    if not isinstance(value, str):
        raise PurposeValidationTypeError

    if value == '':
        return None

    value = ''.join(c for c in value.translate(_PURPOSE_TRANS_TABLE) if c in CHARS_FOR_PURPOSE)
    if len(value) > 210:
        raise PurposeValidationMaxLenError
    return value


def validate_payer_status(value: str) -> str:
    if not isinstance(value, str):
        raise PayerStatusValidationTypeError
    elif value not in (PAYER_STATUSES if date.today().year < CHANGE_YEAR else PAYER_STATUSES_AFTER_2024):
        raise PayerStatusValidationValueError
    return value


def validate_cbc(value: str) -> Optional[str]:
    if not isinstance(value, str):
        raise CBCValidationTypeError
    if value in {'', '0'}:
        return None
    if len(value) != 20:
        raise CBCValidationValueLenError
    elif not only_digits(value):
        raise CBCValidationValueDigitsOnlyError
    if all(c == '0' for c in value):
        raise CBCValidationValueCannotZerosOnly
    return value


def validate_reason(value: str) -> Optional[str]:
    if not isinstance(value, str):
        raise ReasonValidationTypeError
    elif value in {'', '0'}:
        return None
    elif len(value) != 2:
        raise ReasonValidationValueLenError
    return value


def validate_tax_period(value: str) -> Optional[str]:
    if not isinstance(value, str):
        raise TaxPeriodValidationTypeError
    elif value in {'', '0'}:
        return None
    return value


def validate_document_number(value: str) -> Optional[str]:
    if not isinstance(value, str):
        raise DocumentNumberValidationTypeError
    elif value in {'', '0'}:
        return None
    return value


def validate_document_date(value: str) -> Optional[str]:
    if not isinstance(value, str):
        raise DocumentDateValidationTypeError
    elif value in {'', '0'}:
        return None
    return value


def validate_type_of_income(value: str) -> Optional[str]:
    if value in {'', None}:
        return None
    elif not isinstance(value, str):
        raise TypeOfIncomeValidationTypeError
    elif value not in {'1', '2', '3', '4', '5'}:
        raise TypeOfIncomeValidationError
    return value
